import inspect
from typing import Dict, Callable, TypedDict


class Module(TypedDict):
    """
    Modules are blocks of code, representing either object instantiations or function calls. 
    See the mindsdb/lightwood/api/types:Module equivalent for more details.

    :param module: Name of the module (function or class name)
    :param args: Argument to pass to the function or constructor
    """ # noqa
    module: str
    args: Dict[str, str]


def filter_fn_args(fn: Callable, args: Dict) -> Dict:
    filtered = {}
    for k, v in args.items():
        if k in inspect.signature(fn).parameters:
            filtered[k] = v
    return filtered


def is_llm(metric: str) -> bool:
    if metric in ['rag_faithfulness', 'rag_answer_relevancy', 'rag_context_precision', 'rag_context_recall']:
        return True
    else:
        return False




import numpy as np
import pandas as pd


def _setup_rag_eval(data: pd.DataFrame, predictions: pd.Series) -> dict:
    """
    param data: input data for RAGAS
    param predictions: answers from the LLM
    return: dict of the fields in the correct format
    """
    context = list(data['contexts'])
    question = list(data['question'])
    ground_truth = list(data['ground_truth'])
    answer = predictions.tolist()
    return {'contexts': context, 'question': question, 'ground_truth': ground_truth, 'answer': answer}


def evaluate_rag(metric: str,
                 fields: dict) -> float:
    """
    Evaluate the LLM responses using RAGAS.
    param metric: which RAGAS metric to use
    param fields: contains the input fields for RAGAS
    return: score
    """
    try:
        # ugly workaround, see `pyproject.toml` for why
        from datasets import Dataset, Features, Sequence, Value
        from ragas import evaluate
        from ragas.metrics import (
            faithfulness,
            answer_relevancy,
            context_precision,
            context_recall
        )
    except Exception:
        raise Exception("Dependencies for RAG metrics are not available. Re-install using `pip install mindsdb-evaluator[rag] and try again.")  # noqa

    new_context = [[str(entry)] for entry in [fields['contexts']]]

    try:
        data = {
            "question": [fields['question']],
            "contexts": new_context,
            "answer": [fields['answer']],
            "ground_truth": [fields['ground_truth']],
        }
    except KeyError:
        raise Exception(f'''To evaluate `{metric}`, check that your input data contains
                        `question`, `context`, `ground_truth` and try again''')

    # Define the feature types
    features = Features({
        "question": Value("string"),
        "contexts": Sequence(Value("string")),
        "answer": Value("string"),
        "ground_truth": Value("string")
    })

    dataset = Dataset.from_dict(data, features=features)

    # Calculate the score based on the RAGAS metric
    metrics_map = {
        'faithfulness': faithfulness,
        'answer_relevancy': answer_relevancy,
        'context_precision': context_precision,
        'context_recall': context_recall}

    ragas_score = evaluate(dataset, metrics=[metrics_map.get(metric)]).to_pandas()
    result = [score for score in ragas_score[metric] if not np.isnan(score)]
    average = sum(result) / len(result) if result else 0

    return average





import numpy as np
from sklearn.metrics import r2_score


def evaluate_regression_accuracy(
        y_true,
        y_pred,
):
    """
    Evaluates accuracy for regression tasks.
    If predictions have a lower and upper bound, then `within-bound` accuracy is computed: whether the ground truth value falls within the predicted region.
    If not, then a (positive bounded) R2 score is returned instead.

    :return: accuracy score as defined above. 
    """  # noqa
    if 'lower' and 'upper' in y_pred:
        y = np.array(y_true).astype(float)
        within = ((y >= y_pred['lower']) & (y <= y_pred['upper']))
        return sum(within) / len(within)
    else:
        r2 = r2_score(y_true, y_pred['prediction'])
        return max(r2, 0)



from sklearn.metrics import f1_score, top_k_accuracy_score


def evaluate_multilabel_accuracy(y_true, y_pred):
    """
    Evaluates accuracy for multilabel/tag prediction.

    :return: weighted f1 score of y_pred and ground truths.
    """
    pred_values = y_pred['prediction']
    return f1_score(y_true, pred_values, average='weighted')


def evaluate_top_k_accuracy(y_true, y_pred, k: int = 2):
    """
    Evaluates the number of times where the correct label is among the top k labels

    :return: top k accuracy score of y_pred and ground truths.
    """
    pred_values = y_pred['prediction']
    return top_k_accuracy_score(y_true, pred_values, k=k)



from sklearn.metrics import accuracy_score, auc, average_precision_score, balanced_accuracy_score, \
    f1_score, mean_absolute_error, mean_squared_error, mean_squared_log_error, median_absolute_error, \
    mean_absolute_percentage_error, precision_score, r2_score, recall_score, roc_auc_score, silhouette_score, \
    top_k_accuracy_score

from mindsdb_evaluator.accuracy.general import evaluate_accuracy, evaluate_accuracies
from mindsdb_evaluator.accuracy.regression import evaluate_regression_accuracy
from mindsdb_evaluator.accuracy.classification import evaluate_multilabel_accuracy, evaluate_top_k_accuracy
from mindsdb_evaluator.accuracy.forecasting import \
    evaluate_array_accuracy, \
    evaluate_num_array_accuracy, \
    evaluate_cat_array_accuracy, \
    complementary_smape_array_accuracy


__all__ = [
    # internal
    'evaluate_accuracy',
    'evaluate_accuracies',
    'evaluate_regression_accuracy',
    'evaluate_multilabel_accuracy',
    'evaluate_top_k_accuracy',
    'evaluate_array_accuracy',
    'evaluate_num_array_accuracy',
    'evaluate_cat_array_accuracy',
    'complementary_smape_array_accuracy',

    # sklearn
    'accuracy_score',
    'auc',
    'average_precision_score',
    'balanced_accuracy_score',
    'f1_score',
    'mean_absolute_error',
    'mean_squared_error',
    'mean_squared_log_error',
    'median_absolute_error',
    'mean_absolute_percentage_error',
    'precision_score',
    'r2_score',
    'recall_score',
    'roc_auc_score',
    'silhouette_score',
    'top_k_accuracy_score'
]



import importlib
from typing import List, Dict, Optional, Union

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from mindsdb_evaluator.helpers.general import Module, filter_fn_args
from mindsdb_evaluator.accuracy.forecasting import \
    evaluate_array_accuracy, \
    evaluate_num_array_accuracy, \
    evaluate_cat_array_accuracy, \
    complementary_smape_array_accuracy
from mindsdb_evaluator.accuracy.llm_generation import evaluate_rag, _setup_rag_eval


SCORE_TYPES = (float, np.float16, np.float32, np.float64,
               int, np.int8, np.int16, np.int32, np.int64)

if hasattr(np, 'float128'):
    SCORE_TYPES += (np.float128,)


def evaluate_accuracy(data: pd.DataFrame,
                      predictions: pd.Series,
                      accuracy_function: str,
                      target: Optional[str] = None,
                      ts_analysis: Optional[dict] = {},
                      n_decimals: Optional[int] = 3,
                      fn_kwargs: Optional[dict] = {}) -> float:
    """
    Dispatcher for accuracy evaluation.

    :param data: original dataframe.
    :param predictions: output of a machine learning model for the input `data`.
    :param target: target column name.
    :param accuracy_function: either a metric from the `accuracy` module or `scikit-learn.metric`.
    :param ts_analysis: `lightwood.data.timeseries_analyzer` output, used to compute time series task accuracy.
    :param n_decimals: used to round accuracies.
    :param fn_kwargs: additional arguments to be passed to the accuracy function.
    
    :return: accuracy score, given input data and model predictions.
    """  # noqa

    if accuracy_function in ['rag_faithfulness', 'rag_answer_relevancy', 'rag_context_precision', 'rag_context_recall']:
        fields = _setup_rag_eval(data, predictions)
        metric = accuracy_function.replace('rag_', '')
        score = evaluate_rag(metric, fields)
    elif 'array_accuracy' in accuracy_function or accuracy_function in ('bounded_ts_accuracy',):

        if ts_analysis is None or not ts_analysis.get('tss', False) or not ts_analysis['tss'].is_timeseries:
            # normal array, needs to be expanded
            cols = [target]
            y_true = data[cols].apply(lambda x: pd.Series(x[target]), axis=1)
        else:
            horizon = 1 if not isinstance(predictions.iloc[0], list) else len(predictions.iloc[0])
            gby = ts_analysis.get('tss', {}).group_by if ts_analysis.get('tss', {}).group_by else []
            cols = [target] + [f'{target}_timestep_{i}' for i in range(1, horizon)] + gby
            y_true = data[cols]

        y_true = y_true.reset_index(drop=True)
        y_pred = predictions.apply(pd.Series).reset_index(drop=True)  # split array cells into columns

        if accuracy_function == 'evaluate_array_accuracy':
            acc_fn = evaluate_array_accuracy
        elif accuracy_function == 'evaluate_num_array_accuracy':
            acc_fn = evaluate_num_array_accuracy
        elif accuracy_function == 'evaluate_cat_array_accuracy':
            acc_fn = evaluate_cat_array_accuracy
        elif accuracy_function == 'complementary_smape_array_accuracy':
            acc_fn = complementary_smape_array_accuracy
        else:
            raise Exception(f"Could not retrieve accuracy function: {accuracy_function}")

        fn_kwargs['data'] = data[cols]
        fn_kwargs['ts_analysis'] = ts_analysis
        fn_kwargs = filter_fn_args(acc_fn, fn_kwargs)
        score = acc_fn(y_true, y_pred, **fn_kwargs)
    else:
        y_true = data[target].tolist()
        y_pred = list(predictions)
        if hasattr(importlib.import_module('mindsdb_evaluator.accuracy'), accuracy_function):
            accuracy_function = getattr(importlib.import_module('mindsdb_evaluator.accuracy'), accuracy_function)
        else:
            raise Exception(f"Could not retrieve accuracy function: {accuracy_function}")

        try:
            fn_kwargs = filter_fn_args(accuracy_function, fn_kwargs)
            score = accuracy_function(y_true, y_pred, **fn_kwargs)
            assert type(score) in SCORE_TYPES, f"Accuracy function `{accuracy_function.__name__}` returned invalid type {type(score)}"  # noqa
        except ValueError as e:
            if 'mix of label input' in str(e).lower():
                # mixed types, try to convert to string. note: shouldn't happen anymore when labels are passed
                fn_kwargs = filter_fn_args(accuracy_function, fn_kwargs)
                score = accuracy_function([str(y) for y in y_true],
                                          [str(y) for y in y_pred],
                                          **fn_kwargs)
            else:
                raise e

    return round(score, n_decimals)


def evaluate_accuracies(data: pd.DataFrame,
                        predictions: pd.Series,
                        target: str,
                        accuracy_functions: List[Union[str, Module]],
                        ts_analysis: Optional[dict] = {},
                        labels: Optional[list] = [],
                        n_decimals: Optional[int] = 3) -> Dict[str, float]:
    # if a label list is provided, we need to encode the target and predictions accordingly
    if labels:
        le = LabelEncoder()
        le.fit(labels)
        data[target] = le.transform(data[target])
        predictions = le.transform(predictions)

    score_dict = {}
    for accuracy_function in accuracy_functions:
        fn_kwargs = {}
        if not isinstance(accuracy_function, str):
            fn_kwargs = accuracy_function['args']
            accuracy_function = accuracy_function['module']
        score = evaluate_accuracy(data,
                                  predictions,
                                  accuracy_function,
                                  target=target,
                                  ts_analysis=ts_analysis,
                                  n_decimals=n_decimals,
                                  fn_kwargs=fn_kwargs,
                                  )
        score_dict[accuracy_function] = score

    return score_dict



from copy import deepcopy
from typing import Optional, Callable

import numpy as np
import pandas as pd
from sklearn.metrics import r2_score, balanced_accuracy_score


def evaluate_array_accuracy(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        base_acc_fn: Optional[Callable] = None,
) -> float:
    """
    Default forecasting accuracy metric.
    
    Yields mean score over all timesteps in the forecast, as determined by the `base_acc_fn` (R2 score by default).
    """  # noqa

    base_acc_fn = base_acc_fn if base_acc_fn is not None else lambda t, p: max(0, r2_score(t, p))
    fh = y_true.shape[1]
    aggregate = 0.0
    for i in range(fh):
        aggregate += base_acc_fn([t[i] for t in y_true], [p[i] for p in y_pred])

    return aggregate / fh


def evaluate_num_array_accuracy(
        y_true: pd.Series,
        y_pred: pd.Series,
) -> float:
    """
    Numerical forecast accuracy metric.

    Scores are computed for each array index (as determined by the forecast length),
    and the final accuracy is the reciprocal of the average R2 score through all steps.
    """  # noqa

    def _naive(yt: np.ndarray, yp: np.ndarray):
        nan_mask = (~np.isnan(yt)).astype(int)
        yp *= nan_mask
        yt = np.nan_to_num(yt, nan=0.0)
        return evaluate_array_accuracy(yt, yp)

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    return _naive(y_true, y_pred)


def evaluate_cat_array_accuracy(
        y_true: pd.Series,
        y_pred: pd.Series,
        ts_analysis: Optional[dict] = {},
) -> float:
    """
    Categorical forecast accuracy metric.

    Balanced accuracy is computed for each timestep (as determined by the forecast length)
    and the final accuracy is the reciprocal of the average score through all timesteps.
    """  # noqa

    if ts_analysis and ts_analysis['tss'].group_by:
        [y_true.pop(gby_col) for gby_col in ts_analysis['tss'].group_by]

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    return evaluate_array_accuracy(y_true,
                                   y_pred,
                                   base_acc_fn=balanced_accuracy_score)


def complementary_smape_array_accuracy(
        y_true: pd.Series,
        y_pred: pd.Series,
        ts_analysis: Optional[dict] = {},
) -> float:
    """
    Forecast accuracy metric. 
    
    It returns ``1 - (sMAPE/2)``, where ``sMAPE`` is the symmetrical mean absolute percentage error of the forecast versus actual measurements in the time series.

    As such, its domain is 0-1 bounded.
    """  # noqa
    y_true = deepcopy(y_true)
    y_pred = deepcopy(y_pred)
    tss = ts_analysis.get('tss', False)
    if tss and tss.group_by:
        [y_true.pop(gby_col) for gby_col in ts_analysis['tss'].group_by]

    # nan check
    y_true = y_true.values
    y_pred = y_pred.values
    nans = pd.isna(y_true)
    if nans.any():
        # convert all nan indexes to non-zero equal pairs that don't contribute to the metric
        y_true[nans] = 1
        y_pred[nans] = 1

    smape_score = smape(y_true, y_pred)
    csmape = 1 - smape_score / 2
    return csmape


def smape(y_true: np.ndarray, y_pred: np.ndarray):
    """ Symmetrical mean absolute percentage error. """  # noqa
    thres = 1e9
    num = abs(y_pred - y_true)
    den = (abs(y_true) + abs(y_pred)) / 2
    return min(np.average(num / den), thres)
